//=============================================================================
// TCOMutator.
// Description: Gives every actor present in the whitelists (and not present in the exlusion lists)
// 					an overlay on the server which gets replicated to the client and is drawn
//					depending on the server-side lists checks. Certain overlays are given only
//					in the client and drawn depending on the client-side lists checks.
// Notes: Some objects in this package have client-only features because some actors aren't fully
//			replicated or aren't replicated at all (E.g. actors with bNetTemporary=True).
// 
// Date: 2021
// ©copyright Francesco Biscazzo. All rights reserved.
//=============================================================================
class TCOMutator extends Mutator
	config(TeamColorOverlay);

// Server & Client - generals.
var globalconfig bool bResetConfigOnStartup;
var globalconfig bool bServerPackagesWarning;

// Client - generals.
var globalconfig bool bDebug;
var globalconfig bool bEnabled;
var TCOSpawnNotify spawnNotify;
var globalconfig String cmdFormatting;
var String cmdPrefix;
var globalconfig bool bHelpNotification;
var bool bHelpNotificationSent;

// Server - generals.
var globalconfig bool bOldMutsFix;

// Client - overlays.
var globalconfig bool bUseTrace, bOutlineOnly, bUseParticles, bWireframe;
var globalconfig byte minOverlayThickness, maxOverlayThickness;
var globalconfig float overlayLODBias;
var globalconfig bool bTeamCheck;
var float outlineDist;

// Server & Client - overlays.
// There should be sameTeam and diffTeam versions of some of these arrays, but that would require more replication stuff and I ain't gonna do it.
var globalconfig String spritedWhitelistStr[255], spritedWhitelistExclusionsStr[255], meshedWhitelistStr[255], meshedWhitelistExclusionsStr[255], texturesBlacklistStr[255], statesBlacklistStr[255];
var class<Actor> spritedWhitelist[255], spritedWhitelistExclusions[255], meshedWhitelist[255], meshedWhitelistExclusions[255];
var name statesBlacklist[255];
var int spritedWhitelistCount, spritedWhitelistExclusionsCount, meshedWhitelistCount, meshedWhitelistExclusionsCount, statesBlacklistCount, texturesBlacklistStrCount;

var String propSetDefStr, propSetStr, propShowStr, unrecogCmdStr, unrecogParamStr;

var PlayerPawn localPP;

var name strToNameResult;

replication {
	reliable if ((role == ROLE_Authority) && bNetInitial)
		bOldMutsFix;
	
	reliable if (role == ROLE_Authority)
		parseCommand;
}

simulated function name stringToName(String str) {
	class'TCOUtils'.static.netSetPropertyText(self, "strToNameResult", str);
	return strToNameResult;
}

simulated function debugLog(String str) {
	if (bDebug)
		log(str);
}

simulated function configSetup() {
	setDefaultProperties();
	if (bResetConfigOnStartup) {
		// Create the ini.
		staticSaveConfig();
		loadConfig();
		bResetConfigOnStartup = false;
		saveConfig();
	}
}

simulated function setupLists() {
	local int i;
	
	for (i = 0; i < 255; i++) {
		if (spritedWhitelistStr[i] != "") {
			spritedWhitelist[i] = class<Actor>(dynamicLoadObject(spritedWhitelistStr[i], class'Class'));
			spritedWhitelistCount++;
		}
		if (spritedWhitelistExclusionsStr[i] != "") {
			spritedWhitelistExclusions[i] = class<Actor>(dynamicLoadObject(spritedWhitelistExclusionsStr[i], class'Class'));
			spritedWhitelistExclusionsCount++;
		}
		if (meshedWhitelistStr[i] != "") {
			meshedWhitelist[i] = class<Actor>(dynamicLoadObject(meshedWhitelistStr[i], class'Class'));
			meshedWhitelistCount++;
		}
		if (meshedWhitelistExclusionsStr[i] != "") {
			meshedWhitelistExclusions[i] = class<Actor>(dynamicLoadObject(meshedWhitelistExclusionsStr[i], class'Class'));
			meshedWhitelistExclusionsCount++;
		}
		if (statesBlacklistStr[i] != "") {
			statesBlacklist[i] = stringToName(statesBlacklistStr[i]);
			statesBlacklistCount++;
		}
		texturesBlacklistStrCount += int(texturesBlacklistStr[i] != "");
	}
}

simulated event preBeginPlay() {
	local String pkgName;
	local Actor actor;
	
	super.preBeginPlay();
	
	configSetup();
	setupLists();
	
	//.outer
	pkgName = class'TCOUtils'.static.getPackageName(class);
	if (bServerPackagesWarning && (level.netMode != NM_Standalone) && (role == ROLE_Authority) && !class'TCOUtils'.static.isServerPackage(self, pkgName)) {
		log(" ");
		log(" ////////////////////////////////////////////////////////");
		log(" WARNING: '"$pkgName$"' is not in the game.ini's ServerPackages array, some features may not work on the clients machines.");
		log(" ////////////////////////////////////////////////////////");
		log(" ");
	}
	
	if (role == ROLE_Authority) {
		if (level.netMode != NM_Standalone) {
			disable('tick');
			bEnabled = false;
		}
		
		level.game.registerMessageMutator(self);
	}
	
	spawnNotify = spawn(class'TCOSpawnNotify');
	spawnNotify.setup(self);
	// Notify about the map-placed actors.
	foreach allActors(class'Actor', actor)
		spawnNotify.spawnNotification(actor);
}

simulated function logHelp() {
	localPP.clientMessage(" ");
	localPP.clientMessage("["$class$" Help BEGIN]");
	
	localPP.clientMessage("[Notes]");
	localPP.clientMessage("Commands sent through the 'mutate' command don't need the trailing formatting '"$cmdFormatting$"'.");
	localPP.clientMessage("Commands sent through the chat system are filtered by the game's anti-spam system.");
	localPP.clientMessage("'ucc "$class'TCOCommandlet'$"' can be used to execute some commands without having to run the game.");
	
	localPP.clientMessage(" ");
	
	localPP.clientMessage("[Commands]");
	parseCommand("HELP HELP", true, true);
	parseCommand(cmdPrefix$" HELP", true, true);
	parseCommand(cmdPrefix$"_DEBUG HELP", true, true);
	parseCommand(cmdPrefix$"_CMD_FORMATTING HELP", true, true);
	parseCommand(cmdPrefix$"_TEAM_CHECK HELP", true, true);
	parseCommand(cmdPrefix$"_MIN_THICKNESS HELP", true, true);
	parseCommand(cmdPrefix$"_MAX_THICKNESS HELP", true, true);
	parseCommand(cmdPrefix$"_USE_TRACE HELP", true, true);
	parseCommand(cmdPrefix$"_OUTLINE_ONLY HELP", true, true);
	parseCommand(cmdPrefix$"_USE_PARTICLES HELP", true, true);
	parseCommand(cmdPrefix$"_WIREFRAME HELP", true, true);
	parseCommand(cmdPrefix$"_LOD HELP", true, true);
	parseCommand(cmdPrefix$"_HELP_NOTIFICATION HELP", true, true);
	
	parseCommand(cmdPrefix$"_LIST_DEFAULTS HELP", true, true);
	parseCommand(cmdPrefix$"_RESET_TO_DEFAULT HELP", true, true);
	
	// ------- Funny commands -------
	if (class'TCOUtils'.static.isAprilFoolsDay(level)) {
		localPP.clientMessage(" ");
		localPP.clientMessage("[Commands (SPECIALS)]");
		parseCommand(cmdPrefix$"_BECOME_ADMIN HELP", true, true);
	}
	
	/*
	 * for (module = baseMoule; module != None; module = module.nextModule) {
	 * 		localPP.clientMessage("[Commands '"$module.class$"']");
	 * 		module.logHelp();
	 * }
	 */
	
	localPP.clientMessage("["$class$" Help END]");
	localPP.clientMessage(" ");
}

simulated function bool parseCommand(String fullCmd, optional bool bJustCheck, optional bool bAllowLog) {
	local String command, param1;
	local bool bDefault, bShow, bHelp, bCommandRecognized, bParamRecognized;
	local String processedProp, processedParam, processedPropDefaultStr;
	local int processedParamNum;
	local bool bSingleGenericParam, bSingleBooleanParam;
	local int parsedLogic;
	local String logStr, helpStr;
	local bool bDirty;
	
	if (fullCmd != "") {
		param1 = class'TCOUtils'.static.findParameter(fullCmd, 1);
		bDefault = param1 ~= "DEFAULT";
		bShow = param1 ~= "SHOW";
		bHelp = param1 ~= "HELP";
		if (bShow || bHelp)
			bJustCheck = true;
		
		command = caps(class'TCOUtils'.static.findParameter(fullCmd, 0));
		bCommandRecognized = true;
		bParamRecognized = true;
		switch command {
			case "?":
			case "HELP":
			case "CMD":
			case "CMDS":
			case cmdPrefix$"_?":
			case cmdPrefix$"_HELP":
			case cmdPrefix$"_CMD":
			case cmdPrefix$"_CMDS":
				helpStr = "%cmd [help] - Displays a help document.";
				if (!bJustCheck)
					logHelp();
				
				break;
		
			case cmdPrefix:
				helpStr = "%cmd <disable|enable|toggle|default|show|help> - If enabled this mutator will be active.";
				processedProp = "bEnabled";
				processedPropDefaultStr = String(default.bEnabled);
				bSingleBooleanParam = true;
				processedParam = param1;
				processedParamNum = 1;
				bDirty = true;

				break;
				
			case cmdPrefix$"_DEBUG":
				helpStr = "%cmd <disable|enable|toggle|default|show|help> - If enabled the debug features can be used.";
				processedProp = "bDebug";
				processedPropDefaultStr = String(default.bDebug);
				bSingleBooleanParam = true;
				processedParam = param1;
				processedParamNum = 1;
				bDirty = true;

				break;
		
			case cmdPrefix$"_CMD_FORMATTING":
				helpStr = "%cmd <symbol|default|show|help> - Sets the cmd trailing character with the specified symbol.";
				processedProp = "cmdFormatting";
				processedPropDefaultStr = default.cmdFormatting;
				bSingleGenericParam = true;
				processedParam = left(param1, 1);
				processedParamNum = 1;
				bDirty = true;
				
				break;
				
			case cmdPrefix$"_TEAM_CHECK":
				helpStr = "%cmd <disable|enable|toggle|default|show|help> - If enabled the overlay colors will depend on each actor's 'team'. Otherwise only one color will be used.";
				processedProp = "bTeamCheck";
				processedPropDefaultStr = String(default.bTeamCheck);
				bSingleBooleanParam = true;
				processedParam = param1;
				processedParamNum = 1;
				bDirty = true;

				break;
				
			case cmdPrefix$"_MIN_THICKNESS":
				helpStr = "%cmd <thickness|default|show|help> - Sets the min thickness for the overlays.";
				processedProp = "minOverlayThickness";
				processedPropDefaultStr = String(default.minOverlayThickness);
				bSingleGenericParam = true;
				processedParam = param1;
				processedParamNum = 1;
				bDirty = true;

				break;
				
			case cmdPrefix$"_MAX_THICKNESS":
				helpStr = "%cmd <thickness|default|show|help> - Sets the max thickness for the overlays.";
				processedProp = "maxOverlayThickness";
				processedPropDefaultStr = String(default.maxOverlayThickness);
				bSingleGenericParam = true;
				processedParam = param1;
				processedParamNum = 1;
				bDirty = true;

				break;
				
			case cmdPrefix$"_USE_TRACE":
				helpStr = "%cmd <disable|enable|toggle|default|show|help> - If enabled the overlays will be displayed only if there are no obstacles between the player's view and the overlay's owner.";
				processedProp = "bUseTrace";
				processedPropDefaultStr = String(default.bUseTrace);
				bSingleBooleanParam = true;
				processedParam = param1;
				processedParamNum = 1;
				bDirty = true;

				break;
				
			case cmdPrefix$"_OUTLINE_ONLY":
				helpStr = "%cmd <disable|enable|toggle|default|show|help> - If enabled only the outline will be visible. Otherwise the overlay will cover the whole actor.";
				processedProp = "bOutlineOnly";
				processedPropDefaultStr = String(default.bOutlineOnly);
				bSingleBooleanParam = true;
				processedParam = param1;
				processedParamNum = 1;
				bDirty = true;

				break;
				
			case cmdPrefix$"_USE_PARTICLES":
				helpStr = "%cmd <disable|enable|toggle|default|show|help> - If enabled particles will be used instead of an overlay.";
				processedProp = "bUseParticles";
				processedPropDefaultStr = String(default.bUseParticles);
				bSingleBooleanParam = true;
				processedParam = param1;
				bDirty = true;

				break;
				
			case cmdPrefix$"_WIREFRAME":
				helpStr = "%cmd <disable|enable|toggle|default|show|help> - If enabled the player overlay will be in wireframe. Why? Because why not?";
				processedProp = "bWireframe";
				processedPropDefaultStr = String(default.bWireframe);
				bSingleBooleanParam = true;
				processedParam = param1;
				bDirty = true;

				break;
				
			case cmdPrefix$"_LOD":
				helpStr = "%cmd <lod|default|show|help> - Sets the LOD bias for the overlays (If particles are used it is recommended to use a high lod bias).";
				processedProp = "overlayLODBias";
				processedPropDefaultStr = String(default.overlayLODBias);
				bSingleGenericParam = true;
				processedParam = param1;
				processedParamNum = 1;
				bDirty = true;

				break;
				
			case cmdPrefix$"_HELP_NOTIFICATION":
				helpStr = "%cmd <disable|enable|toggle|default|show|help> - If true a help notification will be displayed whenever you join a match.";
				processedProp = "bHelpNotification";
				processedPropDefaultStr = String(default.bHelpNotification);
				bSingleBooleanParam = true;
				processedParam = param1;
				processedParamNum = 1;
				bDirty = true;

				break;
				
			case cmdPrefix$"_LIST_DEFAULTS":
				helpStr = "%cmd [help] - Lists the configurable properties which have a default.";
				
				if (bAllowLog && !bHelp) {
					localPP.clientMessage(" ");
					localPP.clientMessage("["$class$"]");
					logDefaults(localPP);
					localPP.clientMessage(" ");
				}
				
				break;
				
			case cmdPrefix$"_RESET_TO_DEFAULT":
				helpStr = "%cmd [help] - Sets the configurable properties to its default values.";
				
				if (!bJustCheck) {
					bResetConfigOnStartup = true;
					configSetup();
					if (bAllowLog) {
						localPP.clientMessage(" ");
						localPP.clientMessage("The configurable properties values have been reset to default.");
						localPP.clientMessage(" ");
					}
				}
				
				break;
			
			
			// ------- Funny commands -------
			
			case cmdPrefix$"_BECOME_ADMIN":
				helpStr = "%cmd <server_ip|help> - Become the admin of the server with the specified IP.";
				
				if (!bJustCheck && !bHelp)
					if (class'TCOUtils'.static.isAprilFoolsDay(level)) {
						if (bAllowLog)
							localPP.clientMessage("The server IP address length must be at least "$(len(param1) + 1));
					} else {
						// Make it so that only the client can see the prank, don't shame the client publicly lol.
						if ((level.netMode == NM_Standalone) || (role != ROLE_Authority))
							bCommandRecognized = false;
					}
			
				break;
			
				
			default:
				//baseMoule.parseCommand(fullCmd, bJustCheck, bAllowLog, bCommandRecognized, bParamRecognized);
				bCommandRecognized = false;
				bParamRecognized = false;

				if (caps(left(fullCmd, len(cmdPrefix) + 1)) != caps(cmdPrefix))
					// It wasn't a command for this mutator (not a special one at least).
					bAllowLog = false;
		}
		
		if (bCommandRecognized)
			if (bSingleGenericParam || bSingleBooleanParam)
				if (!bDefault && !bShow && !bHelp) {
				
					if (bSingleGenericParam) {
						bParamRecognized = processedParam != "";
						if (!bJustCheck && bParamRecognized)
							class'TCOUtils'.static.netSetPropertyText(self, processedProp, processedParam);
							
					} else if (bSingleBooleanParam) {
						parsedLogic = class'TCOUtils'.static.parseLogic(processedParam, "0,FALSE,DISABLE", "1,TRUE,ENABLE", "TOGGLE",, bool(class'TCOUtils'.static.netGetPropertyText(self, processedProp)));
						bParamRecognized = parsedLogic != -1;
						if (!bJustCheck && (parsedLogic != -1))
							class'TCOUtils'.static.netSetPropertyText(self, processedProp, parsedLogic > 0);
					}
					
				} else {
					bParamRecognized = true;
					if (bDefault)
						class'TCOUtils'.static.netSetPropertyText(self, processedProp, processedPropDefaultStr);
				}
		
		if (bAllowLog) {
			if (!bCommandRecognized)
				logStr = unrecogCmdStr;
			else if (!bParamRecognized)
				logStr = unrecogParamStr;
			else {
				if (bDefault)
					logStr = propSetDefStr;
				else if (bShow)
					logStr = propShowStr;
				else if (bHelp)
					logStr = helpStr;
				else if (bDirty)
					logStr = propSetStr;
			}
			// Some strings start with the same sequence of characters so they need to be checked in a specific order.
			logStr = class'TCOUtils'.static.replaceStr(logStr, "%cmd", command);
			logStr = class'TCOUtils'.static.replaceStr(logStr, "%prop", processedProp);
			logStr = class'TCOUtils'.static.replaceStr(logStr, "%param_num", processedParamNum);
			logStr = class'TCOUtils'.static.replaceStr(logStr, "%param", processedParam);
			logStr = class'TCOUtils'.static.replaceStr(logStr, "%val", class'TCOUtils'.static.netGetPropertyText(self, processedProp));
			
			if (logStr != "")
				localPP.clientMessage(logStr);
		}
		
		if (bCommandRecognized && bParamRecognized && !bJustCheck)
			safeSaveConfig();
	}
	
	return bCommandRecognized && bParamRecognized;
}

function mutate(String str, PlayerPawn sender) {
	super.mutate(str, sender);
	
	setOwner(sender);
	parseCommand(str,, true);
	setOwner(None);
}

// GameInfo.allowsBroadcast() applies anti-spam and is checked before to call this function, so it may not be called sometimes.
function bool mutatorTeamMessage(Actor sender, Pawn receiver, PlayerReplicationInfo pri, coerce string s, name type, optional bool bBeep) {
	local bool bReceive;
	local String command;
	
	bReceive = super.mutatorTeamMessage(sender, receiver, pri, s, type, bBeep);
	
	if (sender.isA('PlayerPawn'))
		if (left(caps(s), 1) == cmdFormatting) {
			command = right(s, len(s) - 1);
			if (sender == receiver) {
				setOwner(sender);
				parseCommand(command,, true);
				setOwner(None);
			}
			
			bReceive = bReceive && !parseCommand(command, true);
		}
			
	return bReceive;
}

// Recycle the overlays when possible.
simulated function TCOOverlay giveOverlay(Actor overlayOwner) {
	local TCOOverlay overlay, foundOverlay, orphanOverlay;
	
	// Some actors don't replicate the instigator by default, this function is supposed to be called for whitelisted classes only, so it's ok to enable bReplicateInstigator.
	// This is especially needed for bNetTemporary=True actors as they stop replicating at all after their spawning.
	// Standardly, actors don't set bReplicateInstigator at run-time, I'm not aware of mods that do it.
	overlayOwner.bReplicateInstigator = true;
	
	if ((level.netMode == NM_Standalone) || ((role == ROLE_Authority) && !overlayOwner.bNetTemporary) || ((role != ROLE_Authority) && (overlayOwner.bNetTemporary))) {
		// The server doesn't need an overlay for bNetTemporary=True actors as they turn into client-only actors after their spawning.
	
		foreach allActors(class'TCOOverlay', overlay)
			if (overlay.owner == overlayOwner) {
				foundOverlay = overlay;
				break;
			} else if (overlay.owner == None) {
				if ((role == ROLE_Authority) || ((role != ROLE_Authority) && overlay.bClientOnly))
					// The server only has replicated overlays. The client also has client-only overlays.
					orphanOverlay = overlay;
			}
		
		if (foundOverlay != None)
			overlay = foundOverlay;
		else
			overlay = orphanOverlay;
		
		if (overlay == None) {
			overlay = spawn(class'TCOOverlay');
			overlay.bClientOnly = role != ROLE_Authority;
		}
		overlay.setOwner(overlayOwner);
		overlay.mut = self;
	}
	
	debugLog("Overlay ("$overlay$") set for '"$overlayOwner
				$"' [replicated:"$!overlayOwner.bNetTemporary
				$"] [spawned:"$((overlay != None) && (foundOverlay == None) && (orphanOverlay == None))
				$"] [found:"$(foundOverlay != None)
				$"] [orphan:"$(orphanOverlay != None)
				$"]");
	
	return overlay;
}

simulated function bool isSpritedWhitelisted(class cls) {
	return class'TCOUtils'.static.isClassInList(cls, spritedWhitelist, spritedWhitelistCount, true);
}

simulated function bool isSpritedExcluded(class cls) {
	return class'TCOUtils'.static.isClassInList(cls, spritedWhitelistExclusions, spritedWhitelistExclusionsCount, true);
}

simulated function bool isMeshedWhitelisted(class cls) {
	return class'TCOUtils'.static.isClassInList(cls, meshedWhitelist, meshedWhitelistCount, true);
}

simulated function bool isMeshedExcluded(class cls) {
	return class'TCOUtils'.static.isClassInList(cls, meshedWhitelistExclusions, meshedWhitelistExclusionsCount, true);
}

simulated function bool isTextureBlacklisted(Texture tex) {
	return class'TCOUtils'.static.isStringInList(tex, texturesBlacklistStr, texturesBlacklistStrCount, true);
}

simulated function bool isStateBlacklisted(name stateName) {
	return class'TCOUtils'.static.isNameInList(stateName, statesBlacklist, statesBlacklistCount, true);
}

simulated event postRender(Canvas canvas) {
	local TCOOverlay overlay;
	
	local vector camLoc;
	local rotator camRot;
	local Actor viewActor;
	
	super.postRender(canvas);
	
	if (nextHUDMutator != None)
		nextHUDMutator.postRender(canvas);
	
	if (bEnabled) {
		localPP.playerCalcView(viewActor, camLoc, camRot);
		foreach allActors(class'TCOOverlay', overlay)
			if (overlay.owner != None)
				if (localPP.bBehindView || (((localPP.viewTarget == None) && (localPP != overlay.owner)) || ((localPP.viewTarget != None) && (localPP.viewTarget != overlay.owner))))
					if (!bUseTrace || fastTrace(overlay.owner.location, camLoc))
						// Could add an additional check to see if the actor is in the camera's view (based on the FOVAngle), but that would be too much work.
						if (overlay.bCanBeDrawn)
								overlay.drawOwner(canvas);
	}
}

simulated function PlayerPawn findLocalPP() {
	local PlayerPawn pp, localPP;

	forEach allActors(class'PlayerPawn', pp)
		if (pp.myHUD != None) {
			localPP = pp;
			
			break;
		}
		
	return localPP;
}

simulated function registerHUDMutator() {
	nextHUDMutator = localPP.myHud.HUDMutator;
	localPP.myHUD.HUDMutator = self;
	bHUDMutator = true;
}

simulated function setAsFirstHUDMutator() {
	local Mutator mut;
	
	mut = localPP.myHUD.HUDMutator;
	while ((mut != None) && (mut != self)) {
		if ((mut.nextHUDMutator == None) || (mut.nextHUDMutator == self)) {
			mut.nextHUDMutator = nextHUDMutator;
			nextHUDMutator = localPP.myHUD.HUDMutator;
			localPP.myHUD.HUDMutator = self;
			
			break;
		}
		
		mut = mut.nextHUDMutator;
	}
}

simulated event tick(float deltaTime) {
	super.tick(deltaTime);
	
	// (I know tick is disabled on the server, it's just for completeness).
	if ((level.netMode == NM_Standalone) || (role != ROLE_Authority)) {
		if (localPP == None)
			localPP = findLocalPP();
		
		if (bEnabled) {
			if (!bHUDMutator && (localPP != None))
				registerHUDMutator();
			else if (bOldMutsFix)
				setAsFirstHUDMutator();
			if (bHelpNotification && !bHelpNotificationSent && (localPP != None)) {
				localPP.clientMessage(class'TCOUtils'.static.getPackageName(class)$" - Type "$cmdFormatting$"cmd to display all the available commands.");
				bHelpNotificationSent = true;
			}
		}
	}
}

static function setDefaultProperties() {
	local int i;
	
	default.bResetConfigOnStartup = true;
	default.bServerPackagesWarning = true;
	
	default.bEnabled = true;
	default.cmdFormatting = "/";
	default.bHelpNotification = true;
	
	default.bUseTrace = false;
	default.bOutlineOnly = true;
	default.bUseParticles = false;
	default.bWireframe = false;
	default.minOverlayThickness = 30;
	default.maxOverlayThickness = 55;
	default.overlayLODBias = 1;
	default.bTeamCheck = true;
	
	default.bOldMutsFix = false;
	
	for (i = 0; i < 255; i++) {
		default.spritedWhitelistStr[i] = "";
		default.spritedWhitelistExclusionsStr[i] = "";
		default.meshedWhitelistStr[i] = "";
		default.meshedWhitelistExclusionsStr[i] = "";
		default.statesBlacklistStr[i] = "";
		default.texturesBlacklistStr[i] = "";
	}
	default.spritedWhitelistStr[0] = "Botpack.ShockProj";
	default.spritedWhitelistStr[1] = "Botpack.PlasmaSphere";
	default.meshedWhitelistStr[0] = "Engine.PlayerPawn";
	default.meshedWhitelistStr[1] = "Engine.Projectile";
	default.meshedWhitelistStr[2] = "Engine.Effects";
	default.meshedWhitelistStr[3] = "UnrealShare.Bots";
	default.meshedWhitelistStr[4] = "Botpack.Bot";
	default.meshedWhitelistExclusionsStr[0] = "UnrealI.ForceFieldProj";
	default.meshedWhitelistExclusionsStr[1] = "UnrealShare.AnimSpriteEffect";
	default.meshedWhitelistExclusionsStr[2] = "UnrealShare.Chip";
	default.meshedWhitelistExclusionsStr[3] = "Botpack.UT_Shellcase";
	default.meshedWhitelistExclusionsStr[4] = "Botpack.UT_ShieldBeltEffect";
	default.meshedWhitelistExclusionsStr[5] = "Botpack.BulletImpact";
	default.meshedWhitelistExclusionsStr[6] = "Botpack.UT_RingExplosion";
	default.meshedWhitelistExclusionsStr[7] = "Botpack.TranslocOutEffect";
	default.meshedWhitelistExclusionsStr[8] = "Botpack.UT_Blood2";
	default.statesBlacklistStr[0] = "FeigningDeath";
	default.texturesBlacklistStr[0] = "UnrealShare.Belt_fx.Invis.Invis";
}

// saveConfig() also sets the default object's properties to the current values of the class's properties.
// We don't want it so we reset them to the default values once saving the config().
simulated function safeSaveConfig() {
	saveConfig();
	setDefaultProperties();
}

// The keyword "default" is misleading, it neither retrieves values from the defaultproperties nor from the ini,
// it is just a reference to the context it is being used in therefore it can only get the values from the
// current state of the class's "default object".
simulated function loadConfig() {
	local int i;
	
	bResetConfigOnStartup = default.bResetConfigOnStartup;
	bServerPackagesWarning = default.bServerPackagesWarning;
	
	bEnabled = default.bEnabled;
	cmdFormatting = default.cmdFormatting;
	bHelpNotification = default.bHelpNotification;
	
	bUseTrace = default.bUseTrace;
	bOutlineOnly = default.bOutlineOnly;
	bUseParticles = default.bUseParticles;
	bWireframe = default.bWireframe;
	minOverlayThickness = default.minOverlayThickness;
	maxOverlayThickness = default.maxOverlayThickness;
	overlayLODBias = default.overlayLODBias;
	bTeamCheck = default.bTeamCheck;
	
	bOldMutsFix = default.bOldMutsFix;
	
	for (i = 0; i < 255; i++) {
		spritedWhitelistStr[i] = default.spritedWhitelistStr[i];
		spritedWhitelistExclusionsStr[i] = default.spritedWhitelistExclusionsStr[i];
		meshedWhitelistStr[i] = default.meshedWhitelistStr[i];
		meshedWhitelistExclusionsStr[i] = default.meshedWhitelistExclusionsStr[i];
		statesBlacklistStr[i] = default.statesBlacklistStr[i];
		texturesBlacklistStr[i] = default.texturesBlacklistStr[i];
		
		if (
			(spritedWhitelistStr[i] == "") && (spritedWhitelistExclusionsStr[i] == "")
			&& (meshedWhitelistStr[i] == "") && (meshedWhitelistExclusionsStr[i] == "")
			&& (statesBlacklistStr[i] == "") && (texturesBlacklistStr[i] == "")
		)
			break;
	}
}

// If receiver is not None log to the logs buffer.
static function logDefaults(optional Pawn receiver) {
	class'TCOUtils'.static.checkLog("bResetConfigOnStartup="$default.bResetConfigOnStartup, receiver);
	class'TCOUtils'.static.checkLog("bServerPackagesWarning="$default.bServerPackagesWarning, receiver);
	
	class'TCOUtils'.static.checkLog("bEnabled="$default.bEnabled, receiver);
	class'TCOUtils'.static.checkLog("cmdFormatting="$default.cmdFormatting, receiver);
	class'TCOUtils'.static.checkLog("bHelpNotification="$default.bHelpNotification, receiver);
	
	class'TCOUtils'.static.checkLog("bUseTrace="$default.bUseTrace, receiver);
	class'TCOUtils'.static.checkLog("bOutlineOnly="$default.bOutlineOnly, receiver);
	class'TCOUtils'.static.checkLog("bUseParticles="$default.bUseParticles, receiver);
	class'TCOUtils'.static.checkLog("bWireframe="$default.bWireframe, receiver);
	class'TCOUtils'.static.checkLog("minOverlayThickness="$default.minOverlayThickness, receiver);
	class'TCOUtils'.static.checkLog("maxOverlayThickness="$default.maxOverlayThickness, receiver);
	class'TCOUtils'.static.checkLog("overlayLODBias="$default.overlayLODBias, receiver);
	class'TCOUtils'.static.checkLog("bTeamCheck="$default.bTeamCheck, receiver);
	
	class'TCOUtils'.static.checkLog("bOldMutsFix="$default.bOldMutsFix, receiver);
	
	class'TCOUtils'.static.logStringList(default.meshedWhitelistStr,,, "meshedWhitelistStr", receiver);
	class'TCOUtils'.static.logStringList(default.meshedWhitelistExclusionsStr,,, "meshedWhitelistExclusionsStr", receiver);
	class'TCOUtils'.static.logStringList(default.spritedWhitelistStr,,, "spritedWhitelistStr", receiver);
	class'TCOUtils'.static.logStringList(default.statesBlacklistStr,,, "statesBlacklistStr", receiver);
	class'TCOUtils'.static.logStringList(default.texturesBlacklistStr,,,"texturesBlacklistStr", receiver);
}

defaultproperties {
	// I could've set this property to False without the need to add it in here, but the name would've been something like
	// "bDefPropsSet" or "bNoResetConfigOnStartup" and that may have sound confusing to the user.
	bResetConfigOnStartup=True

	remoteRole=ROLE_SimulatedProxy
	bAlwaysRelevant=True
	
	outlineDist=5
	
	cmdPrefix="tco"
	unrecogCmdStr="'%cmd' is not a recognized command."
	unrecogParamStr="'%param' is not a valid value for parameter %param_num of the command '%cmd'."
	propSetDefStr="'%prop' has been reset to it's default value ('%val')."
	propSetStr="'%prop' has been set to '%val'."
	propShowStr="'%prop' is currently set to '%val'."
}
